//Title: BeTVOut Class
//Platform: BeOS 5
//Version: 2002.02.19
//Description:
//	A class for controlling BeTVOut-supported hardware.
//

//  Copyright (C) 2002 Kevin H. Patterson

#ifndef _BeTVOut_h
#define _BeTVOut_h

	#include <OS.h>
	#include <Screen.h>

	//For 'BeTVOut' code
		#define VERSION 117	//app version number = 0.11.7
		
		#include "BeTVOut/riva_iic/riva_iic.h"
		#include "BeTVOut/BT868_9.h"
		#include "BeTVOut/Transfer.h"
		#include "BeTVOut/Chrontel.h"
	
	class GraphicsDevice;
	
	class BeTVOut
	{
	friend class GraphicsDevice;
	private:
		BeTVOut(GraphicsDevice * const uGraphicsDevice);
		
	public:
		~BeTVOut();

		bool InitCheck() const;

		status_t TVOut( const	bool uEnable );
		bool TVOut() const;

	public:
		status_t Format( const int32 uFormat );
		int32 Format() const;
		status_t Resolution( const int32 uResolution);
		int32 Resolution() const;
		status_t ColorSpace( const color_space uColorSpace);
		color_space ColorSpace() const;
		status_t OutputAutoDetect( const bool uOutputAutoDetect );
		bool OutputAutoDetect() const;
		status_t Output( const int32 uOutput );
		int32 Output() const;
		status_t Filtering( const int32 uFiltering);
		int32 Filtering() const;
		status_t ColorBars( const bool uColorBars);
		bool ColorBars() const;
		status_t Overscan( const bool uOverscan );
		bool Overscan() const;
		status_t DualDisplay( const bool uDualDisplay );
		bool DualDisplay() const;
		
	private:
		status_t pOutputAutoDetect();
		status_t pOutput();
		status_t pDualDisplay();

	private:
		GraphicsDevice * const iGraphicsDevice;

		bool iInitCheck;
		
		bool iTVOut;
		
		int32 hTVOut;
		//riva_iic device data struct...
		riva_data_t * iTVOutChipData;
		
		int32 iTVOutChipType;
		int	iBTType;		//TVout chiptype
		int iBTVersion;	//TVout chip silicon-version
		byte iCHBus;
		byte iCHAddr;
		byte iCHVersion;

		int32	iOutputDetection;
		
		int32 iFormat;
		int32 iResolution;
		color_space iColorSpace;
		bool iOutputAutoDetect;
		int32 iOutput;
		int32 iFiltering;
		bool iColorBars;
		bool iOverscan;
		bool iDualDisplay;
		
	private:
		static const int32 TVOutChip_Unknown = 0;
		static const int32 TVOutChip_BT = 1;
		static const int32 TVOutChip_CH = 256;

	private:
		int nonVESAsetVESA640(display_mode &new_mode);
		int nonVESAsetVESA800(display_mode &new_mode);
		
	public:
		static const int32 Res_640_480 = 0;
		static const int32 Res_720_480 = 1;
		static const int32 Res_720_576 = 2;
		static const int32 Res_768_576 = 3;
		static const int32 Res_800_600 = 4;
		
		static const int32 Fmt_NTSC = 1;
		static const int32 Fmt_PAL = 4;
		
		static const int32 Out_CVBS = 0x01;
		static const int32 Out_YC = 0x02;
		static const int32 Out_ALL = -1;
	};
	
#endif
