//Title: BeTVOut Class
//Platform: BeOS 5
//Version: 2002.02.19
//Description:
//	A class for controlling BeTVOut-supported hardware.
//
//  Copyright (C) 2002 Kevin H. Patterson

#ifndef _BeTVOut_h
#define _BeTVOut_h

	#include <OS.h>
	#include <Application.h>
	#include <Screen.h>
	#include <Locker.h>
	#include <Autolock.h>

	//For 'BeTVOut' code
		#define VERSION 117	//app version number = 0.11.7
		
		#include "BeTVOut/riva_iic/riva_iic.h"
		#include "BeTVOut/BT868_9.h"
		#include "BeTVOut/Transfer.h"
		#include "BeTVOut/Chrontel.h"
	
	class GraphicsDevice;

	class BeTVOut
	{
	friend class GraphicsDevice;
	private:
		//This class can only be constructed by the GraphicsDevice class.
		BeTVOut(GraphicsDevice * const uGraphicsDevice);
		bool InitCheck() const;
		
	public:
		~BeTVOut();
		
		bool Lock();
		status_t LockWithTimeout( bigtime_t uTimeout );
		void Unlock();

		status_t TVOut( const	bool uEnable );
		bool TVOut() const;

		typedef enum {
			VIDEO_FORMAT_ANY = 0,
			VIDEO_FORMAT_UNKNOWN = 0,
			NTSC = 1,
			NTSC_M = 1,
			NTSC_J = 2,
			PAL_BDGHI = 3,
			PAL = 4,
			PAL_M = 4,
			PAL_N = 5,
			SECAM = 6
		} video_format;
		status_t Format( const video_format uFormat );
		video_format Format() const;

		float RefreshRate() const;
		
		status_t Resolution( const int32 uWidth, const int32 uHeight );
		BRect Frame() const;
		int32 Width() const;
		int32 Height() const;

		status_t ColorSpace( const color_space uColorSpace);
		color_space ColorSpace() const;

		status_t ProposeMode(
			video_format &uFormat,
			float &uFrameRate,
			int32 &uWidth,
			int32 &uHeight,
			color_space &uColorSpace ) const;

		status_t OutputAutoDetect( const bool uOutputAutoDetect );
		bool OutputAutoDetect() const;

		typedef enum {
			OUTPUT_MUX_NONE = 0,
			YC = 0x01,
			S_VIDEO = 0x01,
			CVBS = 0x02,
			COMPOSITE = 0x02,
			OUTPUT_MUX_ALL = -1
		} output_mux;
		status_t Output( const output_mux uOutput );
		output_mux Output() const;

		status_t Filtering( const int32 uFiltering);
		int32 Filtering() const;

		status_t ColorBars( const bool uColorBars);
		bool ColorBars() const;

		status_t Overscan( const bool uOverscan );
		bool Overscan() const;

		typedef enum {
			DISPLAY_MUX_NONE = 0,
			VGA = 0x0001,
			VGA_0 = 0x0001,
			VIDEO = 0x0100,
			VIDEO_0 = 0x0100,
			DISPLAY_MUX_ALL = -1
		} display_mux;
		status_t Display( const display_mux uDualDisplay );
		display_mux Display() const;
		
		status_t Field( int32 &uCurrentField ) const;
		
	private:
		status_t pOutputAutoDetect();
		status_t pOutput();
		status_t pDisplay();

		video_format pSimplifiedFormat() const;
		const unsigned char pBeTVOutMode() const;
		
		int nonVESAsetVESA640(display_mode &new_mode);
		int nonVESAsetVESA800(display_mode &new_mode);

	private:
		//pointer to the "host" graphics device (for hardware control)
		GraphicsDevice * const iGraphicsDevice;
		
		bool iInitCheck;	//Class initialized successfully

		BLocker * iTVOutDeviceMutex;
		
		bool iTVOut;			//TV-Out enabled

		video_format iFormat;
		double iRefreshRate;
		int32 iWidth;
		int32 iHeight;
		color_space iColorSpace;
		bool iOutputAutoDetect;
		output_mux iOutput;
		int32 iFiltering;
		bool iColorBars;
		bool iOverscan;
		display_mux iDisplay;
		
		//Handle to the TV-Out device (for hardware control)
		int32 ihTVOutDevice;

		enum tv_out_device_type {
			TV_OUT_DEVICE_UNKNOWN = 0,
			BT = 1,
			CH = 256
		};
		tv_out_device_type iTVOutDeviceType;
		
		//BETA: The following variables need to be generalized.
			riva_data_t * iTVOutDeviceData;	//riva_iic device data struct
			int	iBTType;			//TVout chiptype
			int iBTVersion;		//TVout chip silicon-version
			byte iCHBus;
			byte iCHAddr;
			byte iCHVersion;

		int32	iOutputDetection;
		
	};
	
#endif
