//Title: Graphics Device Class
//Platform: BeOS 5
//Version: 2001.02.03
//Description:
//	A class for working with physical graphics devices (like VGA displays)
//  Implements a high-level interface to the BeOS graphics drivers and accelerants

//Copyright (C) 2002 Kevin H. Patterson

#ifndef _GraphicsDevice_h
#define _GraphicsDevice_h

	#include <Application.h>
	#include <Locker.h>
	#include <Autolock.h>

	//For 'slimdemo' code
		#include <OS.h>
		#include <image.h>
		#include <FindDirectory.h>
		#include <graphic_driver.h>
		#include <Accelerant.h>
		#include <dirent.h>
		#include <stdio.h>
		#include <string.h>
		#include <sys/stat.h>
		#include <malloc.h>
		#include <errno.h>
		#include <stdlib.h>

	//For 'MediaNode' code
		//#include <Area.h>
		#include <Buffer.h>
		#include <BufferGroup.h>

	class BeTVOut;
	
	class GraphicsDevice
	{
	friend class DeviceManager;
	private:
		GraphicsDevice(const char * DeviceName);
		bool InitCheck();

	public:
		bool Lock();
		status_t LockWithTimeout( bigtime_t uTimeout );
		void Unlock();

		~GraphicsDevice();
		const char * DeviceName();
		const char * Name();
		const char * Chipset();
		const char * SerialNumber();
		uint32 Memory();
		uint32 DACSpeed();

		status_t GetMode( display_mode * uDisplayMode );
		status_t SetMode( display_mode * uDisplayMode, bool uMakeDefault = false );
		status_t ProposeMode(
			display_mode *target,
			const display_mode *low, const display_mode *high );
			
		status_t GetFrameBufferConfig( frame_buffer_config &uFrameBufferConfig ) const;

		status_t MoveDisplayArea(const int x, const int y);
		status_t MoveDisplayArea(const int32 uBufferFlags);

		status_t WaitForRetrace();
		
		BBufferGroup * CreateBufferGroup(const int uDesiredBufferCount = 8);
		void DeleteBufferGroup(BBufferGroup * uBufferGroup);

	private:
		status_t iGetMode();
		
	public:
		BeTVOut * TVOut;

	private:
		image_id LoadAccelerantImage();
		
	private:
		BLocker * iGraphicsDeviceMutex;
	
		const char * iDeviceName;
		bool iInitCheck;
		int hDevice;
		image_id AccelerantImage;
		GetAccelerantHook GetAccelerantFunction;
		accelerant_device_info * AccelerantDeviceInfo;
		
		display_mode iDisplayMode;
		frame_buffer_config iFrameBufferConfig;
		
		bool iFramebufferAreaOwner;
		
		static const int MaxBuffers = 32;
		size_t iBufferOffset[MaxBuffers];
		
		private:
		thread_id volatile iBufferCycler_ThreadID;
		bool volatile iBufferCycler_Run;
		bool volatile iBufferCycler_Idle;

	public:
		status_t BufferCyclerStart();
		status_t BufferCyclerStop( bigtime_t TimeOut = B_INFINITE_TIMEOUT, bool Kill = false );
		status_t WaitForBufferCycle();

	private:
		static int32 iBufferCyclerThread( void * uThreadData );
		bool volatile iWaitingForBufferCycle;
		sem_id volatile iBufferCycleSemaphore;
	};
	
#endif
