// VideoConsumer.cpp

#include <string.h>
#include <Buffer.h>
#include <BufferGroup.h>
#include <TimeSource.h>
#include <ParameterWeb.h>

#include "VideoAddOn.h"
#include "VideoConsumer.h"

#include "BeTVOut.h"

#include "Log.h"
#include <iostream.h>


vint32 VideoConsumer::iInstances = 0;

VideoConsumer::VideoConsumer(
	BMediaAddOn * uAddOn,
	const char * uName,
	int32 uInternalID )
:
	BMediaNode(uName),
	BMediaEventLooper(),
	BBufferConsumer(B_MEDIA_RAW_VIDEO),
	BControllable(),
	
	iInitCheck(B_NO_INIT),
	
	iAddOn(uAddOn),
	iInternalID(uInternalID),
	
	iBufferGroup(NULL),
	
	iLatency(0LL),
	
	iInput(),
	iPreferredFormat(),
	
	iDeviceManager(new DeviceManager),
	iGraphicsDevice(NULL),
	
	iParam_GraphicsDeviceID(0),
	iParam_TVOutDisplayMode(0),
	iParam_TVOutVideoFormat(0),
	iParam_TVOutVideoImage(0),
	iParam_TVOutVideoOutput(0),
	iParam_TVOutVideoFiltering(0.0f),
	
	iParamTime_GraphicsDeviceID(0LL),
	iParamTime_TVOutDisplayMode(0LL),
	iParamTime_TVOutVideoFormat(0LL),
	iParamTime_TVOutVideoImage(0LL),
	iParamTime_TVOutVideoOutput(0LL),
	iParamTime_TVOutVideoFiltering(0LL),
	
	iOutputFramePeriod(0)
{
	Log("VideoConsumer::VideoConsumer\n");
	
	//BETA: Add a check to see if hardware is installed here...
	if (atomic_add(&iInstances, 1) == 0) {
		VideoAddOn::fVideoConsumerPort = ControlPort();
		
		//Initialize our preferred format
		iPreferredFormat.type = B_MEDIA_RAW_VIDEO;
		iPreferredFormat.u.raw_video = media_raw_video_format::wildcard;
		
		//BETA: Some of this information needs to be negotiated with the framebuffer
		//(The lines that have been commented out are left as 'wildcards' (don't care)
		//iPreferredFormat.u.raw_video.field_rate = 29.97;
		iPreferredFormat.u.raw_video.interlace = 1;
		//iPreferredFormat.u.raw_video.first_active = 0;
		//iPreferredFormat.u.raw_video.last_active = 480 - 1;
		iPreferredFormat.u.raw_video.orientation = B_VIDEO_TOP_LEFT_RIGHT;
		
		iPreferredFormat.u.raw_video.display.format = B_RGB32;
		//iPreferredFormat.u.raw_video.display.line_width = 640;
		//iPreferredFormat.u.raw_video.display.line_count = 480;
		//iPreferredFormat.u.raw_video.display.bytes_per_row = 4 * 640;
		
		/* initialize input */
		iInput.source = media_source::null;
		iInput.format = iPreferredFormat;
		
		AddNodeKind(B_PHYSICAL_OUTPUT);
		
		iInitCheck = B_OK;
	}
}


VideoConsumer::~VideoConsumer() {
	Log("VideoConsumer::~VideoConsumer\n");
	if (iInitCheck == B_OK) {
		//HACK: Save configuration settings
		if (iAddOn != NULL)
			((VideoAddOn *) iAddOn)->SaveConfigurationFor(this);
		VideoAddOn::fVideoConsumerPort = 0;
		Quit();
	}
	delete iGraphicsDevice;
	delete iDeviceManager;
	atomic_add(&iInstances, -1);
}


status_t VideoConsumer::InitCheck() const {
	return iInitCheck;
}


status_t VideoConsumer::GetConfiguration( BMessage * outMessage ) const {
	Log("VideoConsumer::GetConfiguration\n");
	
	outMessage->AddInt32("DeviceID", iParam_GraphicsDeviceID);
	outMessage->AddInt32("DisplayMode", iParam_TVOutDisplayMode);
	outMessage->AddInt32("VideoFormat", iParam_TVOutVideoFormat);
	outMessage->AddInt32("VideoImage", iParam_TVOutVideoImage);
	outMessage->AddInt32("VideoOutput", iParam_TVOutVideoOutput);
	outMessage->AddFloat("VideoFiltering", iParam_TVOutVideoFiltering);
	
	return B_OK;
}


status_t VideoConsumer::SetConfiguration( BMessage * inMessage ) {
	Log("VideoConsumer::SetConfiguration\n");
	
	if (inMessage->FindInt32("DeviceID", &iParam_GraphicsDeviceID) != B_OK)
		iParam_GraphicsDeviceID = 0;
	if (inMessage->FindInt32("DisplayMode", &iParam_TVOutDisplayMode) != B_OK)
		iParam_TVOutDisplayMode = 0;
	if (inMessage->FindInt32("VideoFormat", &iParam_TVOutVideoFormat) != B_OK)
		iParam_TVOutVideoFormat = 0;
	if (inMessage->FindInt32("VideoImage", &iParam_TVOutVideoImage) != B_OK)
		iParam_TVOutVideoImage = 0;
	if (inMessage->FindInt32("VideoOutput", &iParam_TVOutVideoOutput) != B_OK)
		iParam_TVOutVideoOutput = 0;
	if (inMessage->FindFloat("VideoFiltering", &iParam_TVOutVideoFiltering) != B_OK)
		iParam_TVOutVideoFiltering = 0.0f;
	
	return B_OK;
}


// BMediaNode

port_id VideoConsumer::ControlPort() const {
	return BMediaNode::ControlPort();
}


BMediaAddOn * VideoConsumer::AddOn( int32 * cookie ) const {
	*cookie = iInternalID;
	return iAddOn;
}


status_t VideoConsumer::HandleMessage(
	int32 message,
	const void * data,
	size_t size )
{
	Log("VideoConsumer::HandleMessage\n");
	// We don't use any private message protocols
	return BBufferConsumer::HandleMessage( message, data, size );
}


status_t VideoConsumer::RequestCompleted( const media_request_info & /* info */ ) {
	 Log("VideoConsumer::RequestCompleted\n");
	 return B_OK;
}


// BMediaEventLooper

void VideoConsumer::NodeRegistered() {
	Log("VideoConsumer::NodeRegistered\n");
	
	if (iInitCheck != B_OK) {
		ReportError( B_NODE_IN_DISTRESS );
	}	else {
		iInput.node = Node();
		iInput.destination.port = ControlPort();
		iInput.destination.id = 0;
		iInput.source = media_source::null;
		iInput.format = iPreferredFormat;
		strncpy( iInput.name, Name(), B_MEDIA_NAME_LENGTH );
		
		SetParameterWeb( CreateParameterWeb() );
		
		BMediaEventLooper::NodeRegistered();
	}
}


void VideoConsumer::Start( bigtime_t performance_time ) {
	BMediaEventLooper::Start( performance_time );
}


void VideoConsumer::Stop( bigtime_t performance_time, bool immediate) {
	BMediaEventLooper::Stop( performance_time, immediate );
}


void VideoConsumer::Seek( bigtime_t media_time, bigtime_t performance_time ) {
	BMediaEventLooper::Seek( media_time, performance_time );
}


void VideoConsumer::TimeWarp( bigtime_t at_real_time, bigtime_t to_performance_time ) {
	BMediaEventLooper::TimeWarp( at_real_time, to_performance_time );
}


void VideoConsumer::HandleEvent(
	const media_timed_event *Event,
	bigtime_t /* Lateness */,
	bool /* RealTimeEvent */ )
{
	switch (Event->type) {
	case BTimedEventQueue::B_HANDLE_BUFFER:
		//Log("VideoConsumer::HandleEvent:Buffer\n");
		{
			BBuffer * Buffer = static_cast<BBuffer *>( Event->pointer );
			if (Buffer != NULL) {
				if (RunState() == BMediaEventLooper::B_STARTED) {
					//BETA: Is this a check to drop late frames?
					if (Event->event_time >= TimeSource()->Now()) {
						if (iInput.source != media_source::null) {
							iGraphicsDevice->WaitForBufferCycle();
							iGraphicsDevice->MoveDisplayArea( Buffer->Flags() );
						}
					}	else
						Log("=Frame Dropped=\n");
				}
				Buffer->Recycle();
			}
			else Log("ERROR: NULL Buffer Received?\n");
		}
		break;

	case BTimedEventQueue::B_START:
		Log("VideoConsumer::HandleEvent:Start\n");
		if (RunState() != BMediaEventLooper::B_STARTED) {
			//BETA: Start the video?
		}
		break;
		
	case BTimedEventQueue::B_STOP:
		Log("VideoConsumer::HandleEvent:Stop\n");
		if (RunState() == BMediaEventLooper::B_STARTED) {
			//BETA: Stop the video stream?
			EventQueue()->FlushEvents( 0,
				BTimedEventQueue::B_ALWAYS, true,
				BTimedEventQueue::B_HANDLE_BUFFER );
		}
		break;
		
	case BTimedEventQueue::B_SEEK:
		Log("VideoConsumer::HandleEvent:Seek\n");
		break;
		
	case BTimedEventQueue::B_WARP:
		Log("VideoConsumer::HandleEvent:Warp\n");
		break;
		
	case BTimedEventQueue::B_PARAMETER:
		Log("VideoConsumer::HandleEvent:Parameter\n");
		break;
		
	}
}


status_t VideoConsumer::DeleteHook( BMediaNode * node ) {
	return BMediaEventLooper::DeleteHook(node);
}


/* BBufferConsumer */

status_t VideoConsumer::AcceptFormat(
	const media_destination & destination,
	media_format * format )
{
	Log("VideoConsumer::AcceptFormat\n");
	
	if (iInput.destination != destination)
		return B_MEDIA_BAD_DESTINATION;
	
	if (format->type != B_MEDIA_RAW_VIDEO)
		return B_MEDIA_BAD_FORMAT;
	
	return SpecializeFormat(format);
}


status_t VideoConsumer::GetNextInput( int32 * cookie, media_input * outInput ) {
	Log("VideoConsumer::GetNextInput\n");
	
	if (*cookie != 0)
		return B_BAD_INDEX;
	
	*outInput = iInput;
	*cookie += 1;
	
	return B_OK;
}


void VideoConsumer::DisposeInputCookie( int32 /* cookie */ ) {
	Log("VideoConsumer::DisposeInputCookie\n");
}


void VideoConsumer::BufferReceived( BBuffer * Buffer ) {
	bigtime_t Now( TimeSource()->Now() );	
  bigtime_t EventTime( Buffer->Header()->start_time );
	//Log( "VideoConsumer::BufferReceived\n" );
  
  //cout << "Buffer Lateness: " << ( EventTime - Now ) << endl;
  
	if (Buffer->Header()->type == B_MEDIA_PARAMETERS) {
		//It's a special Buffer with parameter data
		ApplyParameterData( Buffer->Data(), Buffer->SizeUsed() );
		Buffer->Recycle();

	} else {

		if ( EventTime <= Now ) {
			
			//It's probably an 'always late' buffer from a 'Recording' producer (physical input)
			//Handle this appropriately.
			if (Buffer != NULL) {
				if (RunState() == BMediaEventLooper::B_STARTED) {
					if (iInput.source != media_source::null) {
						
						if ( (TimeSource()->Now() - Buffer->Header()->start_time) < (iOutputFramePeriod + (iOutputFramePeriod >> 1)) ) {
							//bigtime_t tOutputFrameTime( 0 );
							
							iGraphicsDevice->WaitForBufferCycle();
							iGraphicsDevice->MoveDisplayArea( Buffer->Flags() );
							//tOutputFrameTime = TimeSource()->Now();
							
							//cout << "-Frame Latency: " << (tOutputFrameTime - Buffer->Header()->start_time) << endl;
						} else {
							Log("=Frame Dropped=\n");
						}
						
					}
					else Log("ERROR: Buffer From NULL Input?\n");
				}
				else Log("ERROR: Video Consumer not Running!\n");
			}
			else Log("ERROR: NULL Buffer Received?\n");

			Buffer->Recycle();
				
		} else {
	
			//It's a buffer for a future event: Add it to the queue.
			if ( EventQueue()->AddEvent( media_timed_event(
					EventTime,
					BTimedEventQueue::B_HANDLE_BUFFER,
					Buffer,
					BTimedEventQueue::B_RECYCLE_BUFFER) ) < B_OK )
				Buffer->Recycle();

		}
		
	}
}


void VideoConsumer::ProducerDataStatus(
	const media_destination &for_whom,
	int32 status,
	bigtime_t at_performance_time )
{
	Log("VideoConsumer::ProducerDataStatus\n");

	//HACK: Added the following code from the BeBook
	//Is the HandleEvent hook doing anything with these?
	if (for_whom == iInput.destination) {
		media_timed_event event(
			at_performance_time,
			BTimedEventQueue::B_DATA_STATUS, &iInput,
			BTimedEventQueue::B_NO_CLEANUP, status, 0, NULL );
		EventQueue()->AddEvent( event );
	}
}


status_t VideoConsumer::GetLatencyFor(
	const media_destination &for_whom,
	bigtime_t * outLatency,
	media_node_id * outTimeSource )
{
	Log("VideoConsumer::GetLatencyFor\n");
	
	if (for_whom != iInput.destination)
		return B_MEDIA_BAD_DESTINATION;
	
	*outLatency = iLatency;
	*outTimeSource = TimeSource()->ID();
	
	return B_OK;
}


status_t VideoConsumer::Connected(
	const media_source & source,
	const media_destination & destination,
	const media_format & format,
	media_input * outInput )
{
	Log("VideoConsumer::Connected\n");
	
	if (destination != iInput.destination)
		return B_MEDIA_BAD_DESTINATION;

	delete iGraphicsDevice;	
	iGraphicsDevice = iDeviceManager->GetGraphicsDevice( iParam_GraphicsDeviceID );
	if (iGraphicsDevice != NULL) {
		
		//Start TVOut
		if ( iGraphicsDevice->TVOut->
			Format( BeTVOut::video_format(iParam_TVOutVideoFormat) ) < B_OK )
			return B_MEDIA_BAD_FORMAT;
		if ( iGraphicsDevice->TVOut->
			Resolution(
				format.u.raw_video.display.line_width,
				format.u.raw_video.display.line_count ) < B_OK )
			return B_MEDIA_BAD_FORMAT;
			//Of course, we know that its bad form to return B_MEDIA_BAD_FORMAT from
			//this function, but that shouldn't happen since the format should already have
			//been approved by AcceptFormat().
		if ( iGraphicsDevice->TVOut->
			ColorSpace( format.u.raw_video.display.format ) < B_OK )
			return B_MEDIA_BAD_FORMAT;
		iGraphicsDevice->TVOut->OutputAutoDetect( (iParam_TVOutVideoOutput == BeTVOut::OUTPUT_MUX_NONE) );
		iGraphicsDevice->TVOut->Output( BeTVOut::output_mux(iParam_TVOutVideoOutput) );
		iGraphicsDevice->TVOut->Filtering( iParam_TVOutVideoFiltering );
		iGraphicsDevice->TVOut->ColorBars( (iParam_TVOutVideoImage == 2) );
		iGraphicsDevice->TVOut->Overscan( (iParam_TVOutVideoImage == 1) );
		iGraphicsDevice->TVOut->Display( BeTVOut::display_mux(iParam_TVOutDisplayMode) );
		
		iGraphicsDevice->TVOut->TVOut( (iParam_TVOutDisplayMode > 0) );

		iOutputFramePeriod = bigtime_t( 1000000.0f / 30.0f );

		//Get the TVOut Buffer group
		//BETA: Right now we use hard-coded triple buffering (3 buffers), because the BeOS
		//Bt8x8 Video Producer Node (capture device) can't handle more than 3 buffers.
		iBufferGroup = iGraphicsDevice->CreateBufferGroup(3);
		
		//Start the BufferCycler
		iGraphicsDevice->BufferCyclerStart();
		
		/* set the buffers duration */
		if (iBufferGroup != NULL) {
			int32 numBuffers;
			iBufferGroup->CountBuffers(&numBuffers);
			SetBufferDuration(
				(numBuffers * 1000000LL) / format.u.raw_video.field_rate );
		}
		
		//Specify the buffer group for the connection
		int32 userData = 0, changeTag = 0;
		
		if (SetOutputBuffersFor(
			source, destination, iBufferGroup,
		  (void *) &userData, &changeTag, true ) == B_OK)
		  Log("Buffers have been set!\n");

		//Set the node's latency
		//BETA: Our current hard-coded 'nominal' latency is 1.5 frame periods
		iLatency = iOutputFramePeriod + (iOutputFramePeriod >> 1);
		SetEventLatency( iLatency );
		
		//Reserve the connection
		iInput.source = source;
		iInput.format = format;
		
		*outInput = iInput;
		
		return B_OK;
	}
	
	return B_ERROR;
}


void VideoConsumer::Disconnected(
	const media_source & producer,
	const media_destination & where )
{
	Log("VideoConsumer::Disconnected\n");
	
	if (where == iInput.destination && producer == iInput.source) {
		
		if (iGraphicsDevice != NULL) {
		
			iGraphicsDevice->BufferCyclerStop();
			
			iGraphicsDevice->TVOut->ColorBars( true );
			iGraphicsDevice->TVOut->Filtering( false );
			iGraphicsDevice->TVOut->TVOut( false );
		}
		
		//Remove the buffer group for the connection
		int32 userData = 0, changeTag = 0;
		SetOutputBuffersFor(
			producer, where, NULL,
			(void *) &userData, &changeTag, true );
		
		if (iGraphicsDevice != NULL) {
			//Release buffer group
			iGraphicsDevice->DeleteBufferGroup( iBufferGroup );
		}
		delete iGraphicsDevice;
		iGraphicsDevice = NULL;
		
		//Unreserve the connection
		iInput.source = media_source::null;
		iInput.format = iPreferredFormat;
		
	}
}


status_t VideoConsumer::FormatChanged(
	const media_source & producer,
	const media_destination & consumer,
	int32 /* from_change_count */,
	const media_format & /* format */ )
{
	Log("VideoConsumer::FormatChanged\n");
	
	if (consumer != iInput.destination)
		return B_MEDIA_BAD_DESTINATION;
	
	if (producer != iInput.source)
		return B_MEDIA_BAD_SOURCE;
	
	return B_ERROR;
}


// BControllable

status_t VideoConsumer::GetParameterValue(
	int32 id,
	bigtime_t * last_change,
	void * value,
	size_t * size )
{
	switch (id) {
	case Param_DeviceID:
		*last_change = iParamTime_GraphicsDeviceID;
		*((int32 *) value) = iParam_GraphicsDeviceID;
		*size = sizeof(int32);
		break;

	case Param_DisplayMode:
		*last_change = iParamTime_TVOutDisplayMode;
		*((int32 *) value) = iParam_TVOutDisplayMode;
		*size = sizeof(int32);
		break;
	
	case Param_VideoFormat:
		*last_change = iParamTime_TVOutVideoFormat;
		*((int32 *) value) = iParam_TVOutVideoFormat;
		*size = sizeof(int32);
		break;

	case Param_VideoImage:
		*last_change = iParamTime_TVOutVideoImage;
		*((int32 *) value) = iParam_TVOutVideoImage;
		*size = sizeof(int32);
		break;

	case Param_VideoOutput:
		*last_change = iParamTime_TVOutVideoOutput;
		*((int32 *) value) = iParam_TVOutVideoOutput;
		*size = sizeof(int32);
		break;
	
	case Param_VideoFiltering:
		*last_change = iParamTime_TVOutVideoFiltering;
		*((float *) value) = iParam_TVOutVideoFiltering;
		*size = sizeof(float);
		break;
	}
	return B_OK;
}


void VideoConsumer::SetParameterValue(
	int32 id,
	bigtime_t when,
	const void * value,
	size_t /* size */ )
{
	switch (id) {
	case Param_DeviceID:
		iParamTime_GraphicsDeviceID = when;
		iParam_GraphicsDeviceID = *((const int32 *) value);
		BroadcastNewParameterValue(when, id, &iParam_GraphicsDeviceID, sizeof(iParam_GraphicsDeviceID));
		break;
	
	case Param_DisplayMode:
		iParamTime_TVOutDisplayMode = when;
		iParam_TVOutDisplayMode = *((const int32 *) value);
		BroadcastNewParameterValue(when, id, &iParam_TVOutDisplayMode, sizeof(iParam_TVOutDisplayMode));
		if (iGraphicsDevice != NULL)
			iGraphicsDevice->TVOut->Display( BeTVOut::display_mux(iParam_TVOutDisplayMode == 3) );
		break;
	
	case Param_VideoFormat:
		iParamTime_TVOutVideoFormat = when;
		iParam_TVOutVideoFormat = *((const int32 *) value);
		BroadcastNewParameterValue(when, id, &iParam_TVOutVideoFormat, sizeof(iParam_TVOutVideoFormat));
		if (iGraphicsDevice != NULL)
			iGraphicsDevice->TVOut->Format( BeTVOut::video_format(iParam_TVOutVideoFormat) );
		break;
	
	case Param_VideoImage:
		iParamTime_TVOutVideoImage = when;
		iParam_TVOutVideoImage = *((const int32 *) value);
		BroadcastNewParameterValue(when, id, &iParam_TVOutVideoImage, sizeof(iParam_TVOutVideoImage));
		if (iGraphicsDevice != NULL) {
			iGraphicsDevice->TVOut->ColorBars( (iParam_TVOutVideoImage == 2) );
			iGraphicsDevice->TVOut->Overscan( (iParam_TVOutVideoImage == 1) );
		}
		break;
	
	case Param_VideoOutput:
		iParamTime_TVOutVideoOutput = when;
		iParam_TVOutVideoOutput = *((const int32 *) value);
		BroadcastNewParameterValue(when, id, &iParam_TVOutVideoOutput, sizeof(iParam_TVOutVideoOutput));
		if (iGraphicsDevice != NULL) {
			iGraphicsDevice->TVOut->OutputAutoDetect( (iParam_TVOutVideoOutput == 0) );
			iGraphicsDevice->TVOut->Output( BeTVOut::output_mux(iParam_TVOutVideoOutput) );
		}
		break;
	
	case Param_VideoFiltering:
		iParamTime_TVOutVideoFiltering = when;
		iParam_TVOutVideoFiltering = *((const float *) value);
		BroadcastNewParameterValue(when, id, &iParam_TVOutVideoFiltering, sizeof(iParam_TVOutVideoFiltering));
		if (iGraphicsDevice != NULL)
			iGraphicsDevice->TVOut->Filtering( iParam_TVOutVideoFiltering );
		break;
	}
	
	EventQueue()->AddEvent(media_timed_event(when,
		BTimedEventQueue::B_PARAMETER, NULL,
		BTimedEventQueue::B_NO_CLEANUP, id, 0, NULL, 0));
}


status_t VideoConsumer::StartControlPanel( BMessenger * /* messenger */ ) {
	return B_ERROR;
}


// Implementation

status_t VideoConsumer::SpecializeFormat( media_format * format ) {
	if (format == NULL)
		return B_BAD_VALUE;

	if (format->type != B_MEDIA_RAW_VIDEO)
		return B_MEDIA_BAD_FORMAT;
		
	//First, run the format through the TV-Out device's ProposeMode() function.
	if (iDeviceManager == NULL)
		return B_ERROR;
	GraphicsDevice * tGraphicsDevice(NULL);
	tGraphicsDevice = iDeviceManager->GetGraphicsDevice( iParam_GraphicsDeviceID );
	if (tGraphicsDevice == NULL)
		return B_ERROR;
	if (tGraphicsDevice->TVOut == NULL) {
		delete tGraphicsDevice;
		return B_ERROR;
	}

	//Save the current video mode
	display_mode CurrentMode;
	status_t CurrentModeValid;
	CurrentModeValid = tGraphicsDevice->GetMode( &CurrentMode );

	BeTVOut::video_format VideoFormat;
	VideoFormat = BeTVOut::video_format(iParam_TVOutVideoFormat);
	float RefreshRate;
	int32 DisplayWidth, DisplayHeight;
	color_space ColorSpace;

	if (format->u.raw_video.field_rate == media_raw_video_format::wildcard.field_rate)
		RefreshRate = 0;
	else
		RefreshRate = format->u.raw_video.field_rate;

	if (format->u.raw_video.display.line_width == media_raw_video_format::wildcard.display.line_width)
		DisplayWidth = 0;
	else
		DisplayWidth = format->u.raw_video.display.line_width;
	
	if (format->u.raw_video.display.line_count == media_raw_video_format::wildcard.display.line_count)
		DisplayHeight = 0;
	else
		DisplayHeight = format->u.raw_video.display.line_count;
	
	if (format->u.raw_video.display.format == media_raw_video_format::wildcard.display.format)	
		ColorSpace = color_space(0);
	else
		ColorSpace = format->u.raw_video.display.format;
	
	tGraphicsDevice->TVOut->ProposeMode(
		VideoFormat,
		RefreshRate,
		DisplayWidth,
		DisplayHeight,
		ColorSpace );
	
	//Specialize raw video display fields
	if (format->u.raw_video.display.format == media_raw_video_format::wildcard.display.format)
		format->u.raw_video.display.format = ColorSpace;
	if (format->u.raw_video.display.line_width == media_raw_video_format::wildcard.display.line_width)
		format->u.raw_video.display.line_width = DisplayWidth;
	if (format->u.raw_video.display.line_count == media_raw_video_format::wildcard.display.line_count)
		format->u.raw_video.display.line_count = DisplayHeight;
	
	//Specialize raw video fields
	if (format->u.raw_video.field_rate == media_raw_video_format::wildcard.field_rate)
		format->u.raw_video.field_rate = RefreshRate;
	if (format->u.raw_video.interlace == media_raw_video_format::wildcard.interlace)
	    format->u.raw_video.interlace = 1;
	if (format->u.raw_video.first_active == media_raw_video_format::wildcard.first_active)
		format->u.raw_video.first_active = 0;
	if (format->u.raw_video.last_active == media_raw_video_format::wildcard.last_active)
		format->u.raw_video.last_active = DisplayHeight - 1;
	if (format->u.raw_video.orientation == media_raw_video_format::wildcard.orientation)
		format->u.raw_video.orientation = B_VIDEO_TOP_LEFT_RIGHT;
		
	//Clip the first and last active lines
	if (format->u.raw_video.first_active >= format->u.raw_video.display.line_count)
		format->u.raw_video.first_active = format->u.raw_video.display.line_count - 1;
	if (format->u.raw_video.last_active >= format->u.raw_video.display.line_count)
		format->u.raw_video.last_active = format->u.raw_video.display.line_count - 1;
	
	//Clip the field rate
	if (format->u.raw_video.field_rate <= 1.0f)
	    format->u.raw_video.field_rate = 1.0f;
	if (format->u.raw_video.field_rate >= 240.0f)
	    format->u.raw_video.field_rate = 240.0f;

	//Test the mode...
	if ( tGraphicsDevice->TVOut->
		Format( BeTVOut::video_format(iParam_TVOutVideoFormat) ) < B_OK )
	{
		delete tGraphicsDevice;
		return B_MEDIA_BAD_FORMAT;
	}
	if ( tGraphicsDevice->TVOut->
		Resolution(
			format->u.raw_video.display.line_width,
			format->u.raw_video.display.line_count ) < B_OK )
	{
		delete tGraphicsDevice;
		return B_MEDIA_BAD_FORMAT;
	}
	if ( tGraphicsDevice->TVOut->
		ColorSpace( format->u.raw_video.display.format ) < B_OK )
	{
		delete tGraphicsDevice;
		return B_MEDIA_BAD_FORMAT;
	}
	tGraphicsDevice->TVOut->Overscan( (iParam_TVOutVideoImage == 1) );
	tGraphicsDevice->TVOut->Display( BeTVOut::DISPLAY_MUX_NONE );
	if ( tGraphicsDevice->TVOut->TVOut( (iParam_TVOutDisplayMode > 0) ) < B_OK ) {
		delete tGraphicsDevice;
		return B_MEDIA_BAD_FORMAT;
	}

	//Get the TV-Out mode parameters
	frame_buffer_config TVOutConfig;
	tGraphicsDevice->GetFrameBufferConfig( TVOutConfig );
	
	//Disable TV-Out
	tGraphicsDevice->TVOut->TVOut( false );
	
	//Restore the original grahpics mode
	//BETA: HACK: Can't do this right now... SetMode crashes the system depending on the
	//value of some paramters (undefined values?) We can't rely on the values in
	//CurrentMode since we may not be in a 'defined' mode when we get them...
	//I need to clamp all the field values to 'defined' values in SetMode().
	/*if (CurrentModeValid >= B_OK)
		tGraphicsDevice->SetMode( &CurrentMode ); */
	
	delete tGraphicsDevice;
	
	//Forcefully set the video bytes per row field
	format->u.raw_video.display.bytes_per_row = TVOutConfig.bytes_per_row;
	
	return B_OK;
}


BParameterWeb * VideoConsumer::CreateParameterWeb()
{
	BParameterWeb * web = new BParameterWeb();
	
	BParameterGroup * options1 = web->MakeGroup("Device");
	
	BDiscreteParameter * DParam(NULL);
	
	DParam = options1->MakeDiscreteParameter(Param_DeviceID,
		B_MEDIA_NO_TYPE, "Device", B_OUTPUT_MUX);
	
	if (iDeviceManager != NULL) {
		GraphicsDevice * mGraphicsDevice(NULL);
		uint32 mGraphicsDevices = iDeviceManager->GraphicsDeviceCount();
		BString DeviceName;
		for (uint32 idx(0); idx < mGraphicsDevices; ++idx) {
			mGraphicsDevice = iDeviceManager->GetGraphicsDevice(idx);
			if (mGraphicsDevice != NULL) {
				DeviceName = BString("");
				DeviceName << mGraphicsDevice->Name();
				DParam->AddItem( idx, DeviceName.String() );
			}
			delete mGraphicsDevice;
			mGraphicsDevice = NULL;
		}
	}
	
	BParameterGroup *options2 = web->MakeGroup("Output Options");

	BParameterGroup *options20 = options2->MakeGroup("Static");
	
	DParam = options20->MakeDiscreteParameter(Param_DisplayMode,
		B_MEDIA_NO_TYPE, "Display", B_OUTPUT_MUX);
	DParam->AddItem(BeTVOut::VGA, "VGA Only");
	DParam->AddItem(BeTVOut::VIDEO, "Video Out Only");
	DParam->AddItem(BeTVOut::DISPLAY_MUX_ALL, "Both");

	DParam = options20->MakeDiscreteParameter(Param_VideoFormat,
		B_MEDIA_NO_TYPE, "Format", B_VIDEO_FORMAT);
	//BETA: 'Follow Source' Video format needs to be debugged...
	//DParam->AddItem(BeTVOut::VIDEO_FORMAT_UNKNOWN, "Source");
	DParam->AddItem(BeTVOut::NTSC_M, "NTSC-M");
	DParam->AddItem(BeTVOut::NTSC_J, "NTSC-J");
	DParam->AddItem(BeTVOut::PAL_BDGHI, "PAL-BDGHI");
	DParam->AddItem(BeTVOut::PAL_M, "PAL-M");
	DParam->AddItem(BeTVOut::PAL_N, "PAL-N");
	DParam->AddItem(BeTVOut::SECAM, "SECAM");

	DParam = options20->MakeDiscreteParameter(Param_VideoImage,
		B_MEDIA_NO_TYPE, "Image", B_RESOLUTION);
	DParam->AddItem(0, "Center Visible");
	DParam->AddItem(1, "Full Overscan");
	DParam->AddItem(2, "Color Bars");

	BParameterGroup *options21 = options2->MakeGroup("Live");
	
	DParam = options21->MakeDiscreteParameter(Param_VideoOutput,
		B_MEDIA_NO_TYPE, "Output", B_OUTPUT_MUX);
	DParam->AddItem(BeTVOut::OUTPUT_MUX_NONE, "Auto Detect");
	DParam->AddItem(BeTVOut::CVBS, "Composite (CVBS)");
	DParam->AddItem(BeTVOut::YC, "S-Video (Y/C)");

	options21->MakeContinuousParameter(Param_VideoFiltering,
		B_MEDIA_NO_TYPE, "Flicker Filtering", B_LEVEL, "", 0.0f, 3.0f, 1.0f);
	
	return web;
}
